from liara.template import _match_url
import pytest
import liara.site


@pytest.fixture
def default_site():
    site = liara.site.Site()
    return site


def test_match_url(default_site):
    score0 = _match_url('/blog/post/23', '/blog/post/*', default_site)
    assert score0 is not None
    score1 = _match_url('/blog/post/235', '/blog/post/*', default_site)
    assert score1 is not None
    assert score0 < score1


def test_match_url_exact(default_site):
    score = _match_url('/blog', '/blog', default_site)
    assert score == 0
    score = _match_url('/blog/', '/blog/*', default_site)
    assert score > 0


def test_match_fail(default_site):
    score = _match_url('/2014/02/16/2297', '/research/*', default_site)
    assert score is None



import pathlib
import hashlib
import pickle
from typing import Dict
import os


class Cache:
    """A key-value cache."""
    def put(self, key: bytes, value: object) -> bool:
        """Put a value into the cache using the provided key.

        :param key: The key under which ``value`` will be stored.
        :param value: A pickable Python object to be stored.
        :return: ``True`` if the value was added to the cache, ``False`` if
                 it was already cached.
        """
        return True

    def contains(self, key: bytes) -> bool:
        """Check if an object is stored.

        :param key: The key to check.
        :return: ``True`` if such an object exists, else ``False``.
        """
        return False

    def get(self, key: bytes) -> object:
        """Get a stored object.

        :param key: The object key.
        :return: An object if one exists. Otherwise, the behavior is undefined.
                 Use :py:meth:`contains` to check if an object exists.
        """
        return None


class FilesystemCache(Cache):
    """A :py:class:`Cache` implementation which uses the filesystem to cache
    data.

    This cache tries to load a previously generated index. Use
    :py:meth:`persist` to write the cache index to disk.
    """
    __index: Dict[bytes, pathlib.Path]

    def __init__(self, path: pathlib.Path):
        self.__path = path
        self.__index = {}
        os.makedirs(self.__path, exist_ok=True)
        self.__index_file = self.__path / 'cache.index'
        if self.__index_file.exists():
            try:
                self.__index = pickle.load(self.__index_file.open('rb'))
            except Exception:
                # Not being able to load the cache is not an error
                pass

    def persist(self):
        """Persists this cache to disk.

        This function should be called after the cache has been populated. On
        the next run, the constructor will then pick up the index and return
        cached data.
        """
        pickle.dump(self.__index, self.__index_file.open('wb'))

    def put(self, key: bytes, value: object) -> bool:
        if key in self.__index:
            return False

        cache_object_path = self.__path / hashlib.sha256(key).hexdigest()
        pickle.dump(value, cache_object_path.open('wb'))

        self.__index[key] = cache_object_path
        return True

    def contains(self, key: bytes) -> bool:
        return key in self.__index

    def get(self, key: bytes) -> object:
        cache_object_path = self.__index[key]
        return pickle.load(cache_object_path.open('rb'))


class MemoryCache(Cache):
    """An in-memory :py:class:`Cache` implementation.

    This cache stores all objects in-memory.
    """
    __index: Dict[bytes, object]

    def __init__(self):
        self.__index = {}

    def contains(self, key: bytes) -> bool:
        return key in self.__index

    def put(self, key: bytes, value: object) -> bool:
        if key in self.__index:
            return False

        self.__index[key] = object
        return True

    def get(self, key: bytes) -> object:
        return self.__index[key]


import pathlib
import collections.abc


def pairwise(iterable):
    """For a list ``s``, return pairs for consecutive entries. For example,
    a list ``s0``, ``s1``, etc. will produce ``(s0,s1), (s1,s2), ...`` and so
    on.

    See: https://docs.python.org/3/library/itertools.html#recipes."""
    import itertools
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)


def add_suffix(path: pathlib.Path, suffix):
    """Add a suffix to a path.

    This differs from ``with_suffix`` by adding a suffix without changing
    the extension, i.e. adding ``en`` to ``foo.baz`` will produce
    ``foo.en.baz``."""
    name = path.name
    name, sep, extension = name.rpartition('.')
    return path.with_name(name + '.' + suffix + '.' + extension)


def readtime(wordcount: int, words_per_minute=300):
    """Given a number of words, estimate the time it would take to read
    them.

    :return: The time in minutes if it's more than 1, otherwise 1."""
    return max(1, round(wordcount / 300))


def flatten_dictionary(d, sep='.', parent_key=None):
    """Flatten a nested dictionary. This uses the separator to combine keys
    together, so a dictionary access like ``['a']['b']`` with a separator
    ``'.'`` turns into ``'a.b'``."""
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, collections.abc.Mapping):
            items.extend(flatten_dictionary(v, sep=sep,
                                            parent_key=new_key).items())
        else:
            items.append((new_key, v,))
    return dict(items)



from markdown.extensions import Extension
from markdown.treeprocessors import Treeprocessor


class HeadingLevelFixupProcessor(Treeprocessor):
    """This processor demotes headings by one level.

    By default, Markdown starts headings with ``<h1>``, but in general the
    title will be provided by a template. This processor replaces each heading
    with the next-lower heading, and adds a ``demoted`` class.
    """
    def run(self, root):
        return self._demote_header(root)

    def _demote_header(self, element):
        if element.tag == 'h1':
            element.tag = 'h2'
            element.set('class', 'demoted')
        elif element.tag == 'h2':
            element.tag = 'h3'
            element.set('class', 'demoted')
        elif element.tag == 'h3':
            element.tag = 'h4'
            element.set('class', 'demoted')
        elif element.tag == 'h4':
            element.tag = 'h5'
            element.set('class', 'demoted')
        elif element.tag == 'h5':
            element.tag = 'h6'
            element.set('class', 'demoted')
        elif element.tag == 'h6':
            element.tag = 'h6'
            element.set('class', 'demoted')

        for e in element:
            self._demote_header(e)


class HeadingLevelFixupExtension(Extension):
    """Markdown extension for the :py:class:`HeadingLevelFixupProcessor`.
    """
    def extendMarkdown(self, md):
        md.treeprocessors.register(HeadingLevelFixupProcessor(md),
                                   'heading-level-fixup',
                                   100)
        md.registerExtension(self)



from .nodes import (
    DataNode,
    IndexNode,
    DocumentNode,
    ResourceNode,
    Node,
    StaticNode,
    GeneratedNode,
    ThumbnailNode,
)
import pathlib
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
)
from .util import pairwise
import datetime
import logging


def _create_metadata_accessor(field_name):
    """Create an accessor for nested metadata fields.

    This allows accessing attributes or dictionary entries recursively, i.e.
    an accessor ``foo.bar.baz`` will access ``foo['bar']['baz]`` and
    ``foo.bar.baz``, and any combination thereof. Attributes will be preferred
    over dictionary access.
    """
    import operator
    if '.' in field_name:
        def key_fun(o):
            field = field_name.split('.')
            o = o.metadata[field[0]]
            for f in field[1:]:
                if hasattr(o, f):
                    o = getattr(o, f)
                else:
                    o = o[f]
            return o
        return key_fun
    else:
        def key_fun(o):
            return o.metadata[field_name]
        return key_fun


class Collection:
    """A collection is a set of nodes. Collections can be ordered, which allows
    for next/previous queries.
    """

    __log = logging.getLogger('liara.Collection')

    def __init__(self, site, pattern, order_by=[]):
        """
        :param pattern: The pattern to select nodes which belong to this
                        collection.
        :param order_by: A list of accessors for fields to order by. If
                         multiple entries are provided, the result will be
                         sorted by each in order using a stable sort.

        If an ordering is specified, and a particular node cannot support that
        ordering (for instance, as it's missing the field that is used to order
        by), the node will be *removed* from the collection.
        """
        self.__site = site
        self.__filter = pattern
        # We accept a string as well to simplify configuration files
        if isinstance(order_by, str):
            self.__order_by = [order_by]
        else:
            self.__order_by = order_by
        self.__build()

    def __build(self):
        nodes = self.__site.select(self.__filter)

        # We want to remove all nodes that don't have the required fields,
        # as it doesn't make much sense to ask for ordering if some nodes
        # cannot be ordered
        def filter_fun(o):
            for ordering in self.__order_by:
                # we need to match the metadata accessor logic below
                if '.' in ordering:
                    ordering = ordering.split('.')[0]
                if ordering not in o.metadata:
                    return False
            return True

        nodes = filter(filter_fun, nodes)

        for ordering in self.__order_by:
            key_fun = _create_metadata_accessor(ordering)
            nodes = sorted(nodes, key=key_fun)
        pairs = list(pairwise(nodes))
        self.__next = {i[0].path: i[1] for i in pairs}
        self.__previous = {i[1].path: i[0] for i in pairs}
        self.__nodes = {n.path: n for n in nodes}

    @property
    def nodes(self):
        """Get the (sorted) nodes in this collection."""
        return self.__nodes.values()

    def get_next(self, node):
        """Get the next node in this collection with regard to the specified
        order, or ``None`` if this is the last node."""
        if not self.__order_by:
            self.__log.warn('Using get_next() on an unordered collection has '
                            'undefined results')
        return self.__next.get(node.path)

    def get_previous(self, node):
        """Get the previous node in this collection with regard to the specified
        order, or ``None`` if this is the first node."""
        if not self.__order_by:
            self.__log.warn('Using get_next() on an unordered collection has '
                            'undefined results')
        return self.__previous.get(node.path)


def _group_recursive(iterable, group_keys: List[str]):
    import collections
    import itertools
    if not group_keys:
        return list(iterable)

    current_key = group_keys[0]
    group_keys = group_keys[1:]
    if current_key[0] == '*':
        key_func = _create_metadata_accessor(current_key[1:])
        # If the group key starts with a *, this means it's a list of items,
        # not a single item to group by -- for instance, a list of tags
        # In this case, we create one bucket per item, append the items to that
        # list, and eventually we recurse if needed
        result = collections.defaultdict(list)
        for item in iterable:
            groups = key_func(item)
            for group in groups:
                result[group].append(item)

        if group_keys:
            result = {k: _group_recursive(v, group_keys)
                      for k, v in result.items()}
        return result
    else:
        key_func = _create_metadata_accessor(current_key)
        iterable = sorted(iterable, key=key_func)
        result = {}
        for k, v in itertools.groupby(iterable, key_func):
            values = _group_recursive(v, group_keys)
            result[k] = values
        return result


class Index:
    """An index provides an index for a collection, by generating index nodes
    for the collection.
    """
    def __init__(self, site, collection: Collection,
                 path: str, group_by=[], *,
                 create_top_level_index=False):
        nodes = collection.nodes
        self.__groups = _group_recursive(nodes, group_by)
        self.__site = site
        self.__path = path
        self.__create_top_level_index = create_top_level_index

    def create_nodes(self, site):
        """Create the index nodes inside the specified site."""
        self._create_nodes_recursive(site, self.__path,
                                     self.__groups, 1)

        if self.__create_top_level_index:
            url = self.__path
            # Strip off components until we have no parameter left
            while '%' in url:
                url = url[:url.rfind('/')]
            node = IndexNode(pathlib.PurePosixPath(url), {
                'path': self.__path,
                'top_level_index': True
            })
            site.add_index(node)

    def _create_nodes_recursive(self, site, path, d, index):
        for k, v in d.items():
            url = pathlib.PurePosixPath(path.replace(f'%{index}', str(k)))
            # TODO Find out what to do here -- either don't create intermediate
            # nodes (i.e. as long as one %N placeholder is left), or cut off
            # the URL.
            node = IndexNode(url, {'key': k})
            site.add_index(node)
            if isinstance(v, dict):
                self._create_nodes_recursive(site, url, v, index+1)
            else:
                for reference in v:
                    node.add_reference(reference)


class ContentFilter:
    """Content filters can filter out nodes based on various criteria."""
    def apply(self, node: Node) -> bool:
        """Return ``True`` if the node should be kept, and ``False``
        otherwise.
        """
        pass


class DateFilter(ContentFilter):
    """Filter content based on the metadata field ``date``.

    If the date is in the future, the node will be filtered."""
    def __init__(self):
        import tzlocal
        self.__tz = tzlocal.get_localzone()
        self.__now = self.__tz.localize(datetime.datetime.now())

    def apply(self, node: Node) -> bool:
        date = node.metadata.get('date', None)
        if date is None:
            return True

        if date.tzinfo is None:
            date = self.__tz.localize(date)
        return date <= self.__now

    @property
    def reason(self):
        return 'date <= now()'


class StatusFilter(ContentFilter):
    """Filter content based on the metadata field ``status``.

    If ``status`` is set to ``private``, the node will be filtered. The
    comparison is case-insensitive.
    """
    def apply(self, node: Node) -> bool:
        return node.metadata.get('status', '').lower() != 'private'

    @property
    def reason(self):
        return 'status set to "private"'


class ContentFilterFactory:
    def __init__(self):
        self.__filters = {
            'date': DateFilter,
            'status': StatusFilter
        }

    def create_filter(self, name: str) -> ContentFilter:
        class_ = self.__filters[name]
        return class_()


class Site:
    """This class manages to all site content."""

    data: List[DataNode]
    """The list of all data nodes in this site."""

    indices: List[IndexNode]
    """The list of all index nodes in this site."""

    documents: List[DocumentNode]
    """The list of all document nodes in this site."""

    resources: List[ResourceNode]
    """The list of all resources nodes in this site."""

    static: List[StaticNode]
    """The list of all static nodes in this site."""

    generated: List[GeneratedNode]
    """The list of all generated nodes in this site."""
    
    metadata: Dict[str, Any]
    """Metadata describing this site."""

    __nodes: Dict[pathlib.PurePosixPath, Node]
    __root = pathlib.PurePosixPath('/')
    __collections: Dict[str, Collection]
    __indices: List[Index]
    __content_filters: List[ContentFilter]

    def __init__(self):
        self.data = []
        self.indices = []
        self.documents = []
        self.resources = []
        self.static = []
        self.generated = []
        self.metadata = {}
        self.__nodes = {}
        self.__collections = {}
        self.__indices = []
        self.__content_filters = []
        self.__log = logging.getLogger('liara.site')

    def register_content_filter(self, content_filter: ContentFilter):
        """Register a new content filter."""
        self.__content_filters.append(content_filter)

    def set_metadata(self, metadata: Dict[str, Any]) -> None:
        """Set the metadata for this site.

        This overrides any previously set metadata. Metadata is accessible
        via the :py:attr:`metadata` attribute.
        """
        self.metadata = metadata

    def __is_content_filtered(self, node: Node) -> bool:
        for f in self.__content_filters:
            if not f.apply(node):
                self.__log.info(f'Filtered node {node.path} due to {f.reason}')
                return True
        return False

    def add_data(self, node: DataNode) -> None:
        "Add a data node to this site."""
        self.data.append(node)
        self.__register_node(node)

    def add_index(self, node: IndexNode) -> None:
        """Add an index node to this site."""
        self.indices.append(node)
        self.__register_node(node)

    def add_document(self, node: DocumentNode) -> None:
        """Add a document to this site."""
        if self.__is_content_filtered(node):
            return

        self.documents.append(node)
        self.__register_node(node)

    def add_resource(self, node: ResourceNode) -> None:
        """Add a resource to this site."""
        if self.__is_content_filtered(node):
            return

        self.resources.append(node)
        self.__register_node(node)

    def add_static(self, node: StaticNode) -> None:
        """Add a static node to this site."""
        if self.__is_content_filtered(node):
            return

        self.static.append(node)
        self.__register_node(node)

    def add_generated(self, node: GeneratedNode) -> None:
        "Add a generated node to this site."""
        self.generated.append(node)
        self.__register_node(node)

    def __register_node(self, node: Node) -> None:
        if node.path in self.__nodes:
            raise Exception(f'"{node.path}" already exists, cannot overwrite.')
        self.__nodes[node.path] = node

    @property
    def nodes(self) -> Iterable[Node]:
        """The list of all nodes in this site."""
        return self.__nodes.values()

    @property
    def urls(self) -> Iterable[pathlib.PurePosixPath]:
        """The list of all registered URLs."""
        return self.__nodes.keys()

    def create_links(self):
        """This creates links between parents/children.

        This is a separate step so it can be executed after merging nodes
        from multiple sources, for instance themes. It is safe to call this
        function multiple times to create new links; nodes which already
        have a parent are automatically skipped."""
        for key, node in self.__nodes.items():
            if node.parent:
                # The parent is already set, so there's nothing to do
                continue

            parent_path = key.parent
            parent_node = self.__nodes.get(parent_path)
            # The parent_node != node check is required so the root node
            # doesn't get added to itself (by virtue of / being a parent of /)
            if parent_node and parent_node != node:
                parent_node.add_child(node)

    def create_collections(self, collections):
        """Create collections."""
        for name, collection in collections.items():
            order_by = collection.get('order_by', [])
            self.__collections[name] = Collection(self, collection['filter'],
                                                  order_by)

    def create_indices(self, indices):
        """Create indices."""
        for index_definition in indices:
            collection = self.get_collection(index_definition['collection'])
            del index_definition['collection']
            index = Index(self, collection, **index_definition)
            self.__indices.append(index)

        for index in self.__indices:
            index.create_nodes(self)

        # Indices may have been added, so we need to update the links
        self.create_links()

    def create_thumbnails(self, thumbnail_definition):
        """Create thumbnails.

        Based on the thumbnail definition -- which is assumed to be a
        dictionary containing the suffix and the desired size -- this function
        iterates over all static nodes that contain images, and creates new
        thumbnail nodes as required.
        """
        from .util import add_suffix
        new_static = []
        for static in self.static:
            if not static.is_image:
                continue
            static.update_metadata()
            width, height = static.metadata['image_size']
            for k, v in thumbnail_definition.items():
                new_url = add_suffix(static.path, k)
                thumbnail_width = v.get('width', width)
                thumbnail_height = v.get('height', height)
                if width <= thumbnail_width and height <= thumbnail_height:
                    # The image has the right size already (smaller or equal
                    # to the thumbnail), so we just link a new static node
                    copy = StaticNode(static.src, new_url)
                    copy.metadata = static.metadata
                    new_static.append(copy)
                else:
                    thumbnail = ThumbnailNode(static.src, new_url, v)
                    self.add_resource(thumbnail)

        for static in new_static:
            self.add_static(static)

        # Static nodes may have been created, so we need to update the links
        self.create_links()

    def get_next_in_collection(self, 
                               collection: str,
                               node: Node) -> Optional[Node]:
        """Get the next node in a collection."""
        next_node = self.__collections[collection].get_next(node)
        return next_node

    def get_previous_in_collection(self,
                                   collection: str,
                                   node: Node) -> Optional[Node]:
        """Get the previous node in a collection."""
        previous_node = self.__collections[collection].get_previous(node)
        return previous_node

    def get_collection(self, collection: str) -> Collection:
        """Get a collection."""
        return self.__collections[collection]

    def get_node(self, path: pathlib.PurePosixPath) -> Optional[Node]:
        """Get a node based on the URL, or ``None`` if no such node exists."""
        return self.__nodes.get(path)

    def select(self, query: str) -> Iterable[Node]:
        """Select nodes from this site.
        
        The query string may contain ``*`` to list all direct children of a
        node, and ``**`` to recursively enumerate nodes. Partial matches
        using ``*foo`` are not supported."""
        parts = query.split('/')
        # A valid query must start with /
        assert parts[0] == ''
        node = self.__nodes[self.__root]
        for component in parts[1:]:
            if component == '**':
                return node.get_children(recursive=True)
            if component == '*':
                return node.get_children(recursive=False)
            if component == '':
                return [node]

            node = node.get_child(component)
            if node is None:
                return []
        return node



from typing import (
    Dict,
    Any,
)


def create_default_configuration() -> Dict[str, Any]:
    """Creates a dictionary with the default configuration."""
    return {
        'content_directory': 'content',
        'resource_directory': 'resources',
        'static_directory': 'static',
        'output_directory': 'output',
        'build': {
            'clean_output': True,
            'cache_directory': 'cache'
        },
        'content': {
            'filters': ['date', 'status']
        },
        'template': 'templates/default.yaml',
        'routes': {
            'static': 'static_routes.yaml'
        },
        'collections': 'collections.yaml',
        'relaxed_date_parsing': False,
        'allow_relative_links': True,
        'feeds': 'feeds.yaml',
        'metadata': 'metadata.yaml'
    }


def create_default_metadata() -> Dict[str, Any]:
    """Creates a dictionary with the default metadata."""
    return {
        'title': 'Default title',
        'base_url': 'https://example.org',
        'description': 'Default description',
        'language': 'en-US',
        'copyright': 'Licensed under the '
                     '<a href="http://creativecommons.org/licenses/by/4.0/"> '
                     'Creative Commons Attribution 4.0 International License.'
    }



from typing import Dict
from . import Site
import pathlib
from typing import (
    Any,
    Dict,
    Optional,
    Tuple,
)


def _match_url(url: pathlib.PurePosixPath, pattern: str, site: Site) \
        -> Optional[int]:
    """Match an url against a pattern.

    :return: An integer indicating the match score, with 0 being a perfect
             match and higher values being increasingly bad. ``None`` is
             returned if no match was found.
    """
    import fnmatch
    import urllib.parse
    from .nodes import NodeKind
    if '?' in pattern and site:
        pattern, params = pattern.split('?')

        node = site.get_node(url)
        assert node
        params = urllib.parse.parse_qs(params)
        if 'kind' in params:
            kinds = params['kind']
            for kind in kinds:
                if kind == 'document' or kind == 'doc':
                    if node.kind == NodeKind.Document:
                        break
                elif kind == 'index' or kind == 'idx':
                    if node.kind == NodeKind.Index:
                        break
            else:
                return None

    # Exact matches always win
    if pattern == str(url):
        return 0
    # If not exact, we'll look for the longest matching pattern,
    # assuming it is the most specific
    if fnmatch.fnmatch(url, pattern):
        # abs is required, if our pattern is /*, and the url we match against
        # is /, then the pattern is longer than the URL
        return abs(len(str(url)) - len(pattern))

    return None


class Template:
    def render(self, **kwargs):
        pass


class TemplateRepository:
    __definition = Dict[str, str]

    def __init__(self, paths: Dict[str, str]):
        self.__paths = paths

    def update_paths(self, paths: Dict[str, str]):
        self.__paths = paths

    def find_template(self, url: str) -> Template:
        pass

    def _match_template(self, url: str, site: Site) -> str:
        best_match = None
        best_score = None
        for pattern, template in self.__paths.items():
            score = _match_url(url, pattern, site)
            if score is None:
                continue

            if best_score is None or score < best_score:
                best_score = score
                best_match = template

        if not best_match:
            raise Exception(f'Could not find matching template for path: '
                            f'"{url}"')

        return best_match


class MakoTemplate(Template):
    def __init__(self, template):
        self.__template = template

    def render(self, **kwargs) -> str:
        return self.__template.render(**kwargs)


class MakoTemplateRepository(TemplateRepository):
    def __init__(self, paths: Dict[str, str], path: pathlib.Path):
        super().__init__(paths)
        from mako.lookup import TemplateLookup
        self.__lookup = TemplateLookup(directories=[str(path)])

    def find_template(self, url, site: Site) -> Template:
        template = self._match_template(url, site)
        return MakoTemplate(self.__lookup.get_template(template))


class Jinja2Template(Template):
    def __init__(self, template):
        self.__template = template

    def render(self, **kwargs) -> str:
        return self.__template.render(**kwargs)


class Jinja2TemplateRepository(TemplateRepository):
    """Jinja2 based template repository.

    This class has extra magic internally to allow it to be pickled/unpickled,
    which is necessary for multiprocessing."""
    def __init__(self, paths: Dict[str, str], path: pathlib.Path):
        super().__init__(paths)
        self.__path = path
        self.__create_environment()

    def __create_environment(self):
        from jinja2 import FileSystemLoader, Environment
        from .util import readtime
        self.__env = Environment(loader=FileSystemLoader(str(self.__path)))
        self.__env.filters['readtime'] = readtime

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_Jinja2TemplateRepository__env']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.__create_environment()

    def find_template(self, url, site: Site) -> Template:
        template = self._match_template(url, site)
        return Jinja2Template(self.__env.get_template(template))


class Page:
    """A wrapper around :py:class:`DocumentNode` and :py:class:`IndexNode` for
    use inside templates.

    Templates only get applied to those node types, and the :py:class:`Page`
    class provides convenience accessors while hiding the underlying node from
    template code.
    """
    def __init__(self, node):
        self.__node = node

    @property
    def content(self) -> str:
        """Provides the content of this page.
        """
        return self.__node.content

    @property
    def url(self) -> str:
        """Provides the current path of this page.
        """
        # __node.path is a PosixPath object, but inside a template we want to 
        # use a basic string
        return str(self.__node.path)

    @property
    def meta(self) -> Dict[str, Any]:
        """Provides the metadata associated with this page.
        """
        return self.__node.metadata

    @property
    def _node(self):
        return self.__node

    def __str__(self):
        return f'Page({self.url})'

    @property
    def references(self) -> 'query.Query':
        """Provides the list of referenced nodes by this page.

        This can be only used if the current page is an
        :py:class:`~liara.nodes.IndexNode`, in all other cases this will fail.
        For index nodes, this will return the list of references as a
        :py:class:`~liara.query.Query` instance.
        """

        from .nodes import NodeKind
        from .query import Query
        assert self.__node.kind == NodeKind.Index

        return Query(self.__node.references)


class SiteTemplateProxy:
    """A wrapper around :py:class:`Site` for use inside templates.
    """
    __site: Site

    def __init__(self, site: Site):
        self.__site = site
        self.__data = {}
        for data in self.__site.data:
            self.__data.update(data.content)

    @property
    def data(self) -> Dict[str, Any]:
        """Get the union of all :py:class:`liara.nodes.DataNode`
        instances in this site.
        """ 
        return self.__data

    @property
    def metadata(self) -> Dict[str, Any]:
        """Provide access to the metadata of this site.
        """
        return self.__site.metadata

    def select(self, query) -> 'query.Query':
        """Run a query on this site.
        """
        from .query import Query
        return Query(self.__site.select(query))

    def get_previous_in_collection(self, collection: str, page: Page) \
            -> Optional[Page]:
        """Given a collection and a page, return the next page in this
        collection or ``None`` if this is the last page.
        """
        next_node = self.__site.get_previous_in_collection(collection,
                                                           page._node)
        if next_node is not None:
            return Page(next_node)
        else:
            return None

    def get_next_in_collection(self, collection: str, page: Page) \
            -> Optional[Page]:
        """Given a collection and a page, return the previous page in this
        collection or ``None`` if this is the first page.
        """
        previous_node = self.__site.get_next_in_collection(collection,
                                                           page._node)
        if previous_node is not None:
            return Page(previous_node)
        else:
            return None

    def get_collection(self, collection: str) -> 'query.Query':
        """Get a collection in form of a :py:class:`liara.query.Query` for 
        further filtering/sorting.
        """
        from .query import Query
        # We need to turn this into a list as nodes is of type dict_values
        # and can't be sorted/reversed otherwise
        return Query(list(self.__site.get_collection(collection).nodes))



from liara import site, nodes
from collections import namedtuple

item = namedtuple('item', ['metadata', 'name'])


def test_group_recursive():
    items = [
        item({'a': 'key_a_1', 'b': 'key_b_1', 'c': 'key_c_1'}, 1),
        item({'a': 'key_a_1', 'b': 'key_b_1', 'c': 'key_c_1'}, 2),
        item({'a': 'key_a_1', 'b': 'key_b_1', 'c': 'key_c_2'}, 3),
        item({'a': 'key_a_1', 'b': 'key_b_2', 'c': 'key_c_1'}, 4),
        item({'a': 'key_a_1', 'b': 'key_b_1', 'c': 'key_c_1'}, 5),
        item({'a': 'key_a_2', 'b': 'key_b_2', 'c': 'key_c_1'}, 6)
    ]

    groups = site._group_recursive(items, ['a', 'b', 'c'])
    assert len(groups) == 2
    assert 'key_a_1' in groups
    assert 'key_a_2' in groups

    group_1 = groups['key_a_1']
    assert 'key_b_1' in group_1
    assert 'key_b_2' in group_1

    group_1_1 = group_1['key_b_1']
    # c_1 and c_2
    assert len(group_1_1) == 2

    group_1_1_1 = group_1_1['key_c_1']
    # c_1, 1; c_1, 2; c_1, 5
    assert len(group_1_1_1) == 3


def test_group_splat():
    items = [
        item({'tags': ['a', 'b']}, 1),
        item({'tags': ['a', 'c']}, 2),
        item({'tags': ['a', 'd']}, 3),
        item({'tags': ['e', 'b']}, 4),
        item({'tags': ['f', 'b']}, 5),
        item({'tags': ['g', 'b']}, 6)
    ]

    groups = site._group_recursive(items, ['*tags'])
    # 6 tags
    assert len(groups) == 7
    assert len(groups['a']) == 3
    assert items[0] in groups['a']
    assert items[1] in groups['a']
    assert items[2] in groups['a']


class MockDocumentNode(nodes.DocumentNode):
    def __init__(self, path, metadata):
        nodes.Node.__init__(self)

        self.kind = nodes.NodeKind.Document
        self.metadata = metadata
        self.path = path


def test_content_filters():
    s = site.Site()
    cff = site.ContentFilterFactory()

    status_filter = cff.create_filter('status')
    s.register_content_filter(status_filter)

    prvn = MockDocumentNode('/private', {'status': 'private'})
    pubn = MockDocumentNode('/public', {})

    s.add_document(pubn)
    s.add_document(prvn)

    assert len(s.nodes) == 1
    assert '/public' in s.urls


class MockObject:
    def __init__(self, metadata):
        self.metadata = metadata


class NestedMockObject:
    def __init__(self, nested):
        self.nested = nested


def test_metadata_accessor():
    n = NestedMockObject({'bar': 23})
    t = MockObject({'o': n})

    c = site._create_metadata_accessor('o.nested.bar')
    assert c(t) == 23



from liara.nodes import DocumentNode, NodeKind, Node
import pathlib


class MockDocumentNode(DocumentNode):
    def __init__(self, src, metadata):
        self.kind = NodeKind.Document
        self.metadata = metadata
        self.src = src
        self.path = pathlib.PurePosixPath(src)


def test_query_filter_by_tag():
    n1 = MockDocumentNode('/a', {'tags': {'a', 'b'}})
    n2 = MockDocumentNode('/b', {'tags': {'a', 'c'}})
    root = Node()
    root.path = pathlib.PurePosixPath('/')
    root.add_child(n1)
    root.add_child(n2)

    with_a = list(root.select_children().with_tag('a'))
    with_b = list(root.select_children().with_tag('b'))

    assert len(with_a) == 2
    assert len(with_b) == 1



from typing import Iterable, List, Iterator
from .nodes import Node
from .template import Page


class SelectionFilter:
    """Base class for query selection filters."""
    def match(self, node: Node) -> bool:
        """Return ``True`` if the node should be kept, else ``False``."""
        pass


class MetadataFilter(SelectionFilter):
    """Filter items which contain a specific metadata field and optionally
    check if that field matches the provided value.
    """
    def __init__(self, name, value=None):
        self.__name = name
        self.__value = value

    def match(self, node: Node) -> bool:
        if self.__name in node.metadata:
            if self.__value is not None:
                return node.metadata[self.__name] == self.__value
            else:
                return True
        return False


class TagFilter(SelectionFilter):
    """Filter items by a specific tag, this expects a metadata field named
    ``tags`` to be present, and that field must support checks for containment
    using ``in``."""
    def __init__(self, name):
        self.__name = name

    def match(self, node: Node) -> bool:
        tags = node.metadata.get('tags')
        if tags is not None:
            return self.__name in tags
        return False


class Sorter:
    """Base class for query sorters."""
    def __init__(self, reverse=False):
        self._reverse = reverse

    def get_key(self, item):
        """Return the key to be used for sorting."""
        pass

    @property
    def reverse(self):
        return self._reverse


class MetadataSorter(Sorter):
    """Sort nodes by metadata."""
    def __init__(self, item: str, reverse=False, case_sensitive=False):
        super().__init__(reverse)
        self.__item = item
        self.__case_sensitive = case_sensitive

    def get_key(self, item: Page):
        key = item.metadata.get(self.__item)
        if isinstance(key, str) and self.__case_sensitive is False:
            return key.lower()
        return key


class Query(Iterable[Node]):
    """A query modifies a list of nodes, by sorting and filtering entries.
    """

    __filters: List[SelectionFilter]
    __nodes: List[Node]
    __sorters: List[Sorter]
    __limit: int
    __reversed: bool
    __result: List[Page]

    def __init__(self, nodes):
        self.__nodes = nodes
        self.__limit = -1
        self.__filters = []
        self.__sorters = []
        self.__reversed = False
        self.__result = None

    def limit(self, limit) -> 'Query':
        """Limit this query to return at most ``limit`` results."""
        self.__limit = limit
        return self

    def with_metadata(self, name, value=None) -> 'Query':
        """Limit this query to only include nodes which contain the specific
        metadata field.

        :param value: If ``value`` is provided, the field must exist and match
                      the provided value.
        """
        self.__filters.append(MetadataFilter(name, value))
        return self

    def with_tag(self, name) -> 'Query':
        """Limit this query to only include nodes with a metadata field named
        ``tags`` which contains the specified tag name.
        """
        self.__filters.append(TagFilter(name))
        return self

    def sorted_by_title(self, *, reverse=False) -> 'Query':
        """Sort the entries in this query by the metadata field ``title``."""
        return self.sorted_by_metadata('title', reverse=reverse)

    def sorted_by_date(self, *, reverse=False) -> 'Query':
        """Sort the entries in this query by the metadata field ``data``."""
        return self.sorted_by_metadata('date', reverse=reverse)

    def sorted_by_metadata(self, tag: str, *,
                           reverse=False, case_sensitive=False) -> 'Query':
        """Sort the entries in this query by the specified metadata field."""
        self.__sorters.append(MetadataSorter(tag, reverse, case_sensitive))
        return self

    def reversed(self) -> 'Query':
        """Return the results of this query in reversed order."""
        self.__reversed = True
        return self

    def __execute(self):
        if self.__result is not None:
            return

        self.__result = []
        result = self.__nodes
        for f in self.__filters:
            result = filter(lambda x: f.match(x), result)

        if self.__sorters:
            for s in self.__sorters:
                result = sorted(result, key=s.get_key, reverse=s.reverse)

        if self.__reversed:
            result = reversed(result)

        if self.__limit > 0:
            for i, e in enumerate(result):
                if i >= self.__limit:
                    break
                self.__result.append(Page(e))
        else:
            for e in result:
                self.__result.append(Page(e))

    def __iter__(self) -> Iterator[Page]:
        self.__execute()
        return iter(self.__result)

    def __len__(self) -> int:
        self.__execute()
        return len(self.__result)



from .nodes import DocumentNode
from .site import Site
import pathlib
from typing import Dict, List
from enum import Enum, auto
from collections import defaultdict


def _extract_links(document: DocumentNode):
    """Extract all links from ``<a>`` and ``<img>`` tags in a document.

    This assumes the document has been already processed into valid Html.
    """
    from bs4 import BeautifulSoup
    soup = BeautifulSoup(document.content, 'lxml')

    for item in soup.find_all(['img', 'a']):
        if item.name == 'img':
            target = item.attrs.get('src', None)
        elif item.name == 'a':
            target = item.attrs.get('href', None)

        if target and not target.startswith('#'):
            yield target


class LinkType(Enum):
    Internal = auto()
    External = auto()


def _is_internal_link(link: str) -> bool:
    """Check if a link is internal.

    A link is considered internal if it starts with a single slash (not two,
    as this indicates a link using the same protocol.)
    """
    if len(link) >= 2:
        return link[0] == '/' and link[1] != '/'
    else:
        return link[0] == '/'


def gather_links(documents, link_type: LinkType) \
        -> Dict[str, List[pathlib.PurePosixPath]]:
    """Gather links across documents.

    :return: A dictionary containing a link, and the list of document paths
             in which this link was found.
    """
    result = defaultdict(list)

    for document in documents:
        links = _extract_links(document)
        for link in links:
            if link_type == LinkType.Internal and not _is_internal_link(link):
                continue
            elif link_type == LinkType.External and _is_internal_link(link):
                continue

            result[link].append(document.path)

    return result


def validate_internal_links(links: Dict[str, List[pathlib.PurePosixPath]],
                            site: Site):
    """Validate internal links.

    For each link, check if it exists inside the provided site. If not, an
    error is printed indicating the link and the documents referencing it."""

    for link, sources in links.items():
        link = pathlib.PurePosixPath(link)

        # Special case handling for index.html:
        # The redirection table contains full paths including the trailing
        # index.html, so if we find a link `/foo/bar/`, we also check
        # `/foo/bar/index.html` in case a redirection is present.
        index_url = link / 'index.html'
        if index_url in site.urls:
            continue

        if link not in site.urls:
            for source in sources:
                print(f'"{link}" referenced in "{source}" does not exist')


def _check_external_link(url):
    """Issue a request to the external URL and check for a valid response.
    """
    import requests
    ok = False
    try:
        r = requests.get(url, timeout=1)
        if r.status_code == 200:
            ok = True
        else:
            error = f'got {r.status_code}, expected 200'
    except requests.exceptions.ConnectionError:
        error = "connection error"
    except requests.exceptions.ConnectTimeout:
        error = "connection timeout"
    except requests.exceptions.ReadTimeout:
        error = "read timeout"
    except requests.exceptions.TooManyRedirects:
        error = "too many redirects"
    except Exception as e:
        error = str(e)

    if not ok:
        return (False, url, error,)
    else:
        return (True, url, None,)


def validate_external_links(links: Dict[str, List[pathlib.PurePosixPath]]):
    """Validate external links.

    This issues a request for each link, and checks if it connects correctly.
    If not, an error is printed indicating the link and the documents
    referencing it."""
    import multiprocessing

    with multiprocessing.Pool() as pool:
        result = pool.imap_unordered(_check_external_link, links.keys())

        for r in result:
            if not r[0]:
                for source in links[r[1]]:
                    print(f'Link "{r[1]}", referenced in "{source}" failed '
                          f'with: {r[2]}')



import argparse
import cProfile
from . import Liara, __version__
from .config import create_default_configuration
from .yaml import dump_yaml
import logging
import os


def cli():
    """Entry point for the command line."""
    from .nodes import NodeKind
    parser = argparse.ArgumentParser()

    parser.add_argument('--config', default='config.yaml')
    parser.add_argument('--verbose', action='store_true')
    parser.add_argument('--debug', action='store_true')
    parser.add_argument('--version', action='version',
                        version=f'%(prog)s {__version__}')

    subparsers = parser.add_subparsers(title='Commands',
                                       metavar='COMMAND',
                                       help='The command to invoke')

    build_cmd = subparsers.add_parser('build', help='Build the site')
    build_cmd.add_argument('--profile', action='store_true')
    build_cmd.add_argument('--profile-file', default='build.prof')
    build_cmd.set_defaults(func=build)

    validate_links_cmd = subparsers.add_parser('validate-links',
                                               help='Validate links')
    validate_links_cmd.add_argument('--type', choices=['external', 'internal'],
                                    default='internal')
    validate_links_cmd.set_defaults(func=validate_links)

    find_by_tag_cmd = subparsers.add_parser('find-by-tag',
                                            help='Find content by tag')
    find_by_tag_cmd.add_argument('tag', nargs='+')
    find_by_tag_cmd.set_defaults(func=find_by_tag)

    serve_cmd = subparsers.add_parser('serve', help='Start a web server')
    serve_cmd.add_argument('--no-browser', action='store_true',
                           help='Do not open a browser window automatically')
    serve_cmd.set_defaults(func=serve)

    quickstart_cmd = subparsers.add_parser('quickstart',
                                           help='Generate a quickstart blog')
    quickstart_cmd.set_defaults(func=quickstart)

    create_config_cmd = subparsers.add_parser('create-config',
                                              help='Create a default '
                                                   'configuration')
    create_config_cmd.add_argument('-o', '--output', default='config.yaml')
    create_config_cmd.set_defaults(func=create_config)

    list_tags_cmd = subparsers.add_parser('list-tags',
                                          help='List all unique entries in the'
                                               ' metadata.tags field.')
    list_tags_cmd.set_defaults(func=list_tags)

    list_content_cmd = subparsers.add_parser('list-content',
                                             help='Show all tracked content')
    list_content_cmd.add_argument('-f', '--format', choices=['tree', 'list'],
                                  default='tree')
    list_content_cmd.add_argument('-t', '--type', action='append',
                                  choices=list(map(str.lower,
                                               NodeKind.__members__.keys())))
    list_content_cmd.set_defaults(func=list_content)

    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG,
                            format='%(asctime)s %(name)s %(message)s')

        # Unfortunately PIL writes a lot of debug output, so we're disabling it
        # manually here to make debug output useful
        # This doesn't remove critical debug output -- PIL writes things like
        # PNG header data here
        logging.getLogger('PIL').setLevel(logging.INFO)
    elif args.verbose:
        logging.basicConfig(level=logging.INFO,
                            format='%(asctime)s %(name)s %(message)s')
    else:
        logging.basicConfig(level=logging.WARN,
                            format='%(asctime)s %(name)s %(message)s')

    if hasattr(args, 'func'):
        args.func(args)
    else:
        parser.print_help()


def _create_liara(options):
    if os.path.exists(options.config):
        return Liara(options.config)
    else:
        return Liara()


def build(options):
    """Build a site."""
    if options.profile:
        pr = cProfile.Profile()
        pr.enable()
    liara = _create_liara(options)
    liara.build()
    if options.profile:
        pr.disable()
        pr.dump_stats(options.profile_file)


def validate_links(options):
    """Validate links."""
    from .cache import MemoryCache
    from .actions import (
        validate_internal_links,
        validate_external_links,
        gather_links,
        LinkType
    )
    liara = _create_liara(options)
    site = liara.discover_content()
    cache = MemoryCache()

    for document in site.documents:
        document.process(cache)

    if options.type == 'internal':
        link_type = LinkType.Internal
    elif options.type == 'external':
        link_type = LinkType.External

    links = gather_links(site.documents, link_type)

    if options.type == 'internal':
        validate_internal_links(links, site)
    elif options.type == 'external':
        validate_external_links(links)


def list_tags(options):
    """List all tags.

    This uses a metadata field named 'tags' and returns the union of all tags,
    as well as the count how often each tag is used."""
    from collections import Counter
    liara = _create_liara(options)
    site = liara.discover_content()
    tags = []
    for document in site.documents:
        tags += document.metadata.get('tags', [])

    counted_tags = sorted(Counter(tags).items(), key=lambda x: x[1],
                          reverse=True)
    for k, v in counted_tags:
        print(k, v)


def find_by_tag(options):
    """Find pages by tag.

    This searches the metadata for a 'tags' field, which is assumed to be
    a list of tags."""
    liara = _create_liara(options)
    site = liara.discover_content()
    tags = set(options.tag)
    for document in site.documents:
        for tag in document.metadata.get('tags', []):
            if tag in tags:
                print(document.src, tag)
                break


def create_config(options):
    """Create a default configuration.
    
    .. option:: create-config -o <output>"""
    dump_yaml(create_default_configuration(), options.output)


def quickstart(options):
    """Create a quickstart project."""
    from .quickstart import generate
    generate()


def list_content(options):
    """List all content.

    If ``options.format`` is set to ``tree``, this will print the content as a
    tree. If ``options.format`` is set to ``list``, this will produce a
    flat list instead.
    """
    import treelib
    liara = _create_liara(options)
    content = liara.discover_content()

    # We sort by path name, which makes it trivial to sort it later into a tree
    # as children always come after their parent
    nodes = sorted(content.nodes, key=lambda x: x.path)
    if not nodes:
        return

    def filter_nodes(node):
        if node.kind.name.lower() not in options.type:
            return False
        return True

    nodes = list(filter(filter_nodes, nodes))

    if options.format == 'tree':
        tree = treelib.Tree()
        tree.create_node('Site', ('/',))
        if nodes[0].path.parts == ('/',):
            tree.create_node('_index', parent=('/',),
                             data=nodes[0].path)

        known_paths = {('/',)}

        for node in nodes:
            path = node.path
            if len(path.parts) == 1:
                continue
            parent = tuple(path.parts[:-1])

            # The following bit creates intermediate nodes for path segments
            # which are not part of the site tree. For instance, if there's a
            # static file /images/image.jpg, there won't be a /images node. In
            # this case, we create one to allow printing a full tree
            for i in range(len(parent)):
                p = tuple(parent[:i+1])
                if len(p) <= 1:
                    continue
                if p not in known_paths:
                    tree.create_node(f"{parent[i]}", p, parent=tuple(p[:i]),
                                     data=path)
                    known_paths.add(p)

            tree.create_node(
                f"{node.path.parts[-1]} ({node.kind.name})",
                tuple(node.path.parts), parent, data=node.path)
            known_paths.add(tuple(node.path.parts))
        tree.show(key=lambda n: str(n.data).casefold())
    elif options.format == 'list':
        for node in nodes:
            print(str(node.path))


def serve(options):
    """Run a local development server."""
    liara = _create_liara(options)
    liara.serve(open_browser=not options.no_browser)



from .yaml import dump_yaml
from .config import create_default_metadata

__TEMPLATES = {
    'page.jinja2': """
<!DOCTYPE html>
<html>
<head>
    <meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
    <meta name=viewport content="width=device-width, initial-scale=1">
      <title>
      {% block title %}
          {% if page.meta.title and page.url != '/' %}
            {{ page.meta.title }}
        {% else %}
            My blog
        {% endif %}
        {% endblock %}
    </title>
    <link type="text/css" rel="stylesheet" href="/style.css" />
</head>
<body>
    <div class="container">
        <header id="header">
            <nav>
                <a href="/">Home</a>
                <a href="/archive">Archive</a>
            </nav>
        </header>
        <main id="content">
        {% block content %}
            {{ page.content }}
        {% endblock %}
        </main>
        <footer id="footer">
            <p>Generated using Liara</p>
        </footer>
    </div>
</body>
</html>
    """,
    'blog.jinja2': """{% extends "page.jinja2" %}
{% block content %}
<article>
    <header>
        <h1>{{page.meta.title}}</h1>
        <div class="metadata">
            <div class="post-time">
              <time datetime="{{ page.meta.date }}">{{
                  page.meta.date.strftime("%Y-%m-%d, %H:%M") }}</time>
            </div>
            <div class="tags">
              <ul class="tags">
                {% for tag in page.meta.tags|sort %}
                  <li>
                    <a href="/archive/by-tag/{{ tag }}">{{ tag|capitalize }}</a>
                   </li>
                {% endfor %}
              </ul>
            </div>
        </div>
    </header>
    {{ page.content }}
    <aside>
        <div class="pagination">
            {% set previous = site.get_previous_in_collection('blog', page) %}
            {% set next = site.get_next_in_collection('blog', page) %}
            {% if previous %}
                <div class="previous">
                    <a href="{{ previous.url }}">Previous post</a>
                </div>
            {% endif %}
            {% if next %}
                <div class="next">
                    <a href="{{ next.url }}">Next post</a>
                </div>
            {% endif %}
        </div>
    </aside>
</article>
{% endblock %}
    """,
    'archive.jinja2': """
{% extends "page.jinja2" %}
{% block content %}
<h1>Blog archive {% if page.meta.key %}for {{ page.meta.key }}{% endif %}</h1>
{{ page.content }}
<ul>
    {% if page.references %}
        {% for ref in page.references %}
        <li>
            <a href="{{ ref.url }}">{{ ref.meta.title }}</a>
        </li>
        {% endfor %}
    {% else %}
        {% for page in
        node.select_children().sorted_by_metadata('key') %}
        <li>
            <a href="{{ page.url }}">{{ page.meta.key }}</a>
            ({{ page.references|length }} posts)
            <ul>
                {% for ref in page.references.sorted_by_date(reverse=True) %}
                <li>
                    <a href="{{ ref.url }}">{{ ref.meta.title}}</a>
                </li>
                {% endfor %}
            </ul>
        </li>
        {% endfor %}
    {% endif %}
</ul>
{% endblock %}
    """
}

__SCSS = """
.container {
    max-width: 800px;

    margin: auto;

    header {
        background-color: #EEE;
        padding: 0.5em;
    }

    main {
        padding: 0.5em;

        header {
            padding: 0;
            background-color: unset;
        }
    }

    footer {
        font-size: 0.8em;
        padding: 0.5em;
        border-top: 1px solid #CCC;
    }
}
"""

__THEME_CONFIG = {
    'backend': 'jinja2',
    'paths': {
        '/blog/*?kind=document': 'blog.jinja2',
        '/archive/*': 'archive.jinja2',
        '/*': 'page.jinja2',
    },
    'resource_directory': 'resources'
}


def generate_templates():
    for k, v in __TEMPLATES.items():
        open(f'templates/{k}', 'w').write(v)


def generate_css():
    import os
    os.makedirs('templates/resources', exist_ok=True)
    open('templates/resources/style.scss', 'w').write(__SCSS)


def generate_theme():
    import os
    os.makedirs('templates', exist_ok=True)
    generate_templates()
    generate_css()
    dump_yaml(__THEME_CONFIG, open('templates/default.yaml', 'w'))


def generate_content():
    import datetime
    import os
    page_template = """---
title: %TITLE%
tags: [%TAGS%]
date: %DATE%
---

Sample post using _Markdown_
"""

    dates = [
        datetime.datetime(2019, 1, 3),
        datetime.datetime(2018, 1, 5),
        datetime.datetime(2017, 3, 1)
    ]

    titles = [
        'Newest post',
        'Old but not oldest post',
        'The beginning'
    ]

    tags = [
        ['new', 'featured'],
        ['old'],
        ['old', 'historic']
    ]

    slugs = [
        'newest-post',
        'old-but-not-oldest-post',
        'the-beginning'
    ]

    os.makedirs('content/blog', exist_ok=True)

    for date, title, tag, slug in zip(dates, titles, tags, slugs):
        t = page_template.replace("%TITLE%", title)
        t = t.replace("%TAGS%", ', '.join(tag))
        t = t.replace("%DATE%", str(date))
        os.makedirs(f'content/blog/{date.year}', exist_ok=True)
        open(f'content/blog/{date.year}/{slug}.md', 'w').write(t)

    open('content/archive.md', 'w').write("""
---
title: The archive
---

The blog archive. Find blog posts [by year](/archive/by-year) or
[by tag](/archive/by-tag).""")

    open('content/_index.md', 'w').write("""
---
title: Hello world
---

Welcome to the sample blog.""")


__DEFAULT_CONFIG = {
    'collections': 'collections.yaml',
    'indices': 'indices.yaml'
}

__DEFAULT_COLLECTIONS = {
    'blog': {
        'filter': '/blog/**',
        'order_by': 'date'
    }
}

__DEFAULT_INDICES = [
    {
        'collection': 'blog',
        'group_by': ['date.year'],
        'path': '/archive/by-year/%1',
        'create_top_level_index': True
    },
    {
        'collection': 'blog',
        'group_by': ['*tags'],
        'path': '/archive/by-tag/%1',
        'create_top_level_index': True
    }
]


def generate_configs():
    dump_yaml(__DEFAULT_CONFIG, open('config.yaml', 'w'))
    dump_yaml(create_default_metadata(), open('metadata.yaml', 'w'))
    dump_yaml(__DEFAULT_COLLECTIONS, open('collections.yaml', 'w'))
    dump_yaml(__DEFAULT_INDICES, open('indices.yaml', 'w'))


def generate():
    generate_theme()
    generate_content()
    generate_configs()



from liara.util import add_suffix, flatten_dictionary, pairwise


def test_flatten_dictionary():
    d = {'a': 1, 'b': {'c': 2}}
    x = flatten_dictionary(d)
    assert x == {'a': 1, 'b.c': 2}


def test_add_suffix():
    import pathlib
    p = pathlib.Path('foo.bar')
    p = add_suffix(p, 'baz')

    assert str(p) == 'foo.baz.bar'


def test_pairwise():
    assert list(pairwise([1, 2, 3])) == [(1, 2,), (2, 3,)]


def test_pairwise_one_element_list():
    assert list(pairwise([1])) == []


def test_pairwise_empty_list():
    assert list(pairwise([])) == []



from liara.nodes import extract_metadata_content
import pytest


def test_extract_toml_metadata():
    document = """+++
a = "b"
+++

content
"""
    metadata, content = extract_metadata_content(document)
    assert 'a' in metadata
    assert metadata['a'] == 'b'
    assert content == """
content
"""


def test_extract_yaml_metadata():
    document = """---
a: "b"
---

content
"""
    metadata, content = extract_metadata_content(document)
    assert 'a' in metadata
    assert metadata['a'] == 'b'
    assert content == """
content
"""


def test_extract_metadata_mismatch_throws():
    document = """---
a = "b"
+++

content
"""

    with pytest.raises(Exception):
        m, c = extract_metadata_content(document)


def test_extract_no_metadata():
    document = """
content
"""

    metadata, content = extract_metadata_content(document)
    assert content == """
content
"""
    assert metadata is None



import os
import pathlib
from typing import (
        List,
        Callable,
    )
import collections
from .yaml import load_yaml
from .site import Site, ContentFilterFactory
from .nodes import (
    DocumentNodeFactory,
    RedirectionNode,
    ResourceNodeFactory,
)
from .cache import Cache, FilesystemCache
from .util import flatten_dictionary
import logging
from . import config


__version__ = '2.0a7'
__all__ = [
    'actions',
    'cache',
    'cmdline',
    'config',
    'feeds',
    'md',
    'nodes',
    'publish',
    'query',
    'quickstart',
    'server',
    'site',
    'template',
    'util',
    'yaml'
]


__ROOT_PATH = pathlib.PurePosixPath('/')


# Create the path from the full path as discovered during walk
# This turns something like 'directory/foo/bar' into '/foo/bar'
def _create_relative_path(path: pathlib.Path, root: pathlib.Path) \
        -> pathlib.PurePosixPath:
    # Extra check, as with_name would fail on an empty path
    if path == root:
        return __ROOT_PATH

    path = path.relative_to(root)
    return __ROOT_PATH / pathlib.PurePosixPath(path.with_name(path.stem))


class Liara:
    __site: Site
    __resource_node_factory: ResourceNodeFactory
    __document_node_factory: DocumentNodeFactory
    __redirections: List[RedirectionNode]
    __log = logging.getLogger('liara')
    __cache: Cache
    __document_post_processors = List[Callable]

    def __init__(self, configuration=None, *, configuration_overrides={}):
        self.__site = Site()
        self.__redirections = []

        default_configuration = config.create_default_configuration()
        if configuration is None:
            project_configuration = {}
        elif isinstance(configuration, str):
            project_configuration = load_yaml(open(configuration))
        else:
            project_configuration = load_yaml(configuration)
        self.__configuration = collections.ChainMap(
            # Must be flattened already
            configuration_overrides,
            flatten_dictionary(project_configuration),
            flatten_dictionary(default_configuration))

        self.__resource_node_factory = ResourceNodeFactory()
        self.__document_node_factory = DocumentNodeFactory(
            self.__configuration
        )

        template_configuration = pathlib.Path(self.__configuration['template'])
        self.__setup_template_backend(template_configuration)

        cache_directory = pathlib.Path(
            self.__configuration['build.cache_directory'])
        self.__cache = FilesystemCache(cache_directory)

        self.__setup_content_filters(self.__configuration['content.filters'])

    def __setup_content_filters(self, filters: List[str]) -> None:
        content_filter_factory = ContentFilterFactory()
        for f in filters:
            self.__site.register_content_filter(
                content_filter_factory.create_filter(f)
            )

    def __setup_template_backend(self, configuration_file: pathlib.Path):
        from .template import Jinja2TemplateRepository, MakoTemplateRepository

        template_path = configuration_file.parent
        configuration = load_yaml(open(configuration_file))

        backend = configuration['backend']
        paths = configuration['paths']

        self.__thumbnail_definition = configuration.get(
            'image_thumbnail_sizes', {})

        if backend == 'jinja2':
            self.__template_repository = Jinja2TemplateRepository(
                paths, template_path)
        elif backend == 'mako':
            self.__template_repository = MakoTemplateRepository(
                paths, template_path)
        else:
            raise Exception(f'Unknown template backend: "{backend}"')

        if 'resource_directory' in configuration:
            resource_directory = pathlib.Path(
                configuration['resource_directory'])

            self.__discover_resources(self.__site,
                                      self.__resource_node_factory,
                                      template_path / resource_directory)

        if 'static_directory' in configuration:
            static_directory = pathlib.Path(
                configuration['static_directory'])
            self.__discover_static(self.__site,
                                   template_path / static_directory)

    def __discover_redirections(self, site: Site, static_routes: pathlib.Path):
        from .nodes import RedirectionNode
        if not static_routes.exists():
            return

        routes = load_yaml(static_routes.open())
        for route in routes:
            node = RedirectionNode(
                    pathlib.PurePosixPath(route['src']),
                    pathlib.PurePosixPath(route['dst']))
            self.__redirections.append(node)
            site.add_generated(node)

    def __discover_content(self, site: Site, content_root: pathlib.Path) \
            -> None:
        from .nodes import DataNode, IndexNode, Node, StaticNode

        document_factory = self.__document_node_factory

        for (dirpath, _, filenames) in os.walk(content_root):
            directory = pathlib.Path(dirpath)
            # Need to run two passes here: First, we check if an _index file is
            # present in this folder, in which case it's the root of this
            # directory
            # Otherwise, we create a new index node
            node: Node
            indexNode: Node = None
            for filename in filenames:
                if filename.startswith('_index'):
                    src = pathlib.Path(os.path.join(dirpath, filename))
                    relative_path = _create_relative_path(directory,
                                                          content_root)
                    node = document_factory.create_node(src.suffix,
                                                        src, relative_path)
                    site.add_document(node)
                    break
            else:
                node = IndexNode(_create_relative_path(directory,
                                                       content_root))
                site.add_index(node)
                indexNode = node

            for filename in filenames:
                if filename.startswith('_index'):
                    continue

                src = pathlib.Path(os.path.join(directory, filename))
                path = _create_relative_path(src, content_root)

                if src.suffix in document_factory.known_types:
                    node = document_factory.create_node(src.suffix, src, path)
                    site.add_document(node)
                    # If there's an index node, we add each document directly
                    # below it manually to the reference list
                    # This way, a simply query using index.references returns
                    # all documents, instead of having to go through the
                    # children and filter by type
                    if indexNode:
                        indexNode.add_reference(node)
                elif src.suffix in {'.yaml'}:
                    node = DataNode(src, path)
                    site.add_data(node)
                else:
                    metadata_path = src.with_suffix('.meta')
                    path = path.with_suffix(''.join(src.suffixes))
                    if metadata_path.exists():
                        node = StaticNode(src, path, metadata_path)
                    else:
                        node = StaticNode(src, path)
                    site.add_static(node)

    def __discover_static(self, site: Site, static_root: pathlib.Path) -> None:
        from .nodes import StaticNode
        for dirpath, _, filenames in os.walk(static_root):
            directory = pathlib.Path(dirpath)

            for filename in filenames:
                src = directory / filename
                path = _create_relative_path(src, static_root)
                # We need to re-append the source suffix
                # We can't use .with_suffix, as this will break on paths like
                # a.b.c, where with_suffix('foo') will produce a.b.foo instead
                # of a.b.c.foo
                path = path.parent / (path.name + src.suffix)

                metadata_path = src.with_suffix('.meta')
                if metadata_path.exists():
                    node = StaticNode(src, path, metadata_path)
                else:
                    node = StaticNode(src, path)
                site.add_static(node)

    def __discover_resources(self, site: Site,
                             resource_factory: ResourceNodeFactory,
                             resource_root: pathlib.Path) -> None:
        for dirpath, _, filenames in os.walk(resource_root):
            for filename in filenames:
                src = pathlib.Path(os.path.join(dirpath, filename))
                path = _create_relative_path(src, resource_root)

                metadata_path = src.with_suffix('.meta')
                if metadata_path.exists():
                    node = resource_factory.create_node(src.suffix, src, path,
                                                        metadata_path)
                else:
                    node = resource_factory.create_node(src.suffix, src, path)
                site.add_resource(node)

    def __discover_feeds(self, site: Site, feeds: pathlib.Path) -> None:
        from .feeds import JsonFeedNode, RSSFeedNode, SitemapXmlFeedNode

        if not feeds.exists():
            return

        for key, options in load_yaml(feeds.open()).items():
            path = pathlib.PurePosixPath(options['path'])
            del options['path']

            if key == 'rss':
                feed = RSSFeedNode(path, site, **options)
                site.add_generated(feed)
            elif key == 'json':
                feed = JsonFeedNode(path, site, **options)
                site.add_generated(feed)
            elif key == 'sitemap':
                feed = SitemapXmlFeedNode(path, site, **options)
                site.add_generated(feed)
            else:
                self.__log.warn(f'Unknown feed type: "{key}", ignored')

    def __discover_metadata(self, site: Site, metadata: pathlib.Path) -> None:
        if not metadata.exists():
            return

        site.set_metadata(load_yaml(metadata.open()))

    def discover_content(self) -> Site:
        self.__log.info('Discovering content ...')
        configuration = self.__configuration

        metadata = pathlib.Path(configuration['metadata'])
        self.__discover_metadata(self.__site, metadata)

        content_root = pathlib.Path(configuration['content_directory'])
        self.__discover_content(self.__site, content_root)

        static_root = pathlib.Path(configuration['static_directory'])
        self.__discover_static(self.__site, static_root)

        resource_root = pathlib.Path(configuration['resource_directory'])
        self.__discover_resources(self.__site, self.__resource_node_factory,
                                  resource_root)

        static_routes = pathlib.Path(configuration['routes.static'])
        self.__discover_redirections(self.__site, static_routes)

        # Feeds use metadata, so this must come after self.__discover_metadata
        feeds = pathlib.Path(configuration['feeds'])
        self.__discover_feeds(self.__site, feeds)

        if self.__thumbnail_definition:
            self.__site.create_thumbnails(self.__thumbnail_definition)

        self.__site.create_links()

        if 'collections' in configuration:
            collections = pathlib.Path(configuration['collections'])
            if collections.exists():
                self.__site.create_collections(
                    load_yaml(collections.read_text()))

        if 'indices' in configuration:
            indices = pathlib.Path(configuration['indices'])
            if indices.exists():
                self.__site.create_indices(
                    load_yaml(indices.read_text()))

        self.__site.create_links()

        self.__log.info(f'Discovered {len(self.__site.nodes)} items')

        return self.__site

    @property
    def site(self) -> Site:
        return self.__site

    def __clean_output(self):
        import shutil
        output_directory = self.__configuration['output_directory']
        self.__log.info(f'Cleaning output directory: "{output_directory}" ...')
        if os.path.exists(output_directory):
            shutil.rmtree(output_directory)
        self.__log.info('Output directory cleaned')

    def build(self):
        from .publish import TemplatePublisher
        self.__log.info('Build started')
        if self.__configuration['build.clean_output']:
            self.__clean_output()

        site = self.discover_content()

        for document in site.documents:
            document.validate_metadata()

        self.__log.info('Processing documents ...')
        for document in site.documents:
            document.process(self.__cache)
        self.__log.info(f'Processed {len(site.documents)} documents')

        self.__log.info('Processing resources ...')
        for resource in site.resources:
            resource.process(self.__cache)
        self.__log.info(f'Processed {len(site.resources)} resources')

        output_path = pathlib.Path(self.__configuration['output_directory'])

        publisher = TemplatePublisher(output_path, site,
                                      self.__template_repository)

        self.__log.info('Publishing ...')
        for document in site.documents:
            document.publish(publisher)
        self.__log.info(f'Published {len(site.documents)} document(s)')

        for index in site.indices:
            index.publish(publisher)
        self.__log.info(f'Published {len(site.indices)} '
                        f'{"indices" if len(site.indices)>1 else "index"}')

        for resource in site.resources:
            resource.publish(publisher)
        self.__log.info(f'Published {len(site.resources)} resource(s)')

        for static in site.static:
            static.publish(publisher)
        self.__log.info(f'Published {len(site.static)} static file(s)')

        if site.generated:
            for generated in site.generated:
                generated.generate()
                generated.publish(publisher)
            self.__log.info(f'Published {len(site.generated)} '
                            'generated file(s)')

        if self.__redirections:
            self.__log.info('Writing redirection file ...')
            with (output_path / '.htaccess').open('w') as output:
                for node in self.__redirections:
                    output.write(f'RedirectPermanent {str(node.path)} '
                                 f'{str(node.dst)}\n')
            self.__log.info(f'Wrote {len(self.__redirections)} redirections')

        self.__log.info('Build finished')
        self.__cache.persist()

    def serve(self, *, open_browser=True):
        from .server import HttpServer
        if self.__configuration['build.clean_output']:
            self.__clean_output()

        site = self.discover_content()

        for document in site.documents:
            document.validate_metadata()

        server = HttpServer(site, self.__template_repository,
                            self.__configuration,
                            open_browser=open_browser)
        server.serve()



from .nodes import (
    DocumentNode,
    IndexNode,
    Publisher,
    ResourceNode,
    StaticNode,
    GeneratedNode,
)
from .template import Page, SiteTemplateProxy, TemplateRepository
from .site import Site
import pathlib
from typing import (
    Union
)


def _publish_with_template(output_path: pathlib.Path,
                           node: Union[DocumentNode, IndexNode],
                           site: Site,
                           site_template_proxy: SiteTemplateProxy,
                           template_repository) -> pathlib.Path:
    page = Page(node)
    file_path = pathlib.Path(str(output_path) + str(node.path))
    file_path.mkdir(parents=True, exist_ok=True)
    file_path = file_path / 'index.html'

    template = template_repository.find_template(node.path, site)
    file_path.write_text(template.render(
        site=site_template_proxy,
        page=page,
        node=node), encoding='utf-8')

    return file_path


class DefaultPublisher(Publisher):
    def __init__(self, output_path: pathlib.Path,
                 site: Site):
        self._output_path = output_path
        self._site = site

    def publish_resource(self, resource: ResourceNode):
        import os
        assert resource is not None
        file_path = pathlib.Path(str(self._output_path) + str(resource.path))
        os.makedirs(file_path.parent, exist_ok=True)
        file_path.write_bytes(resource.content)
        return file_path

    def publish_generated(self, generated: GeneratedNode):
        import os
        file_path = pathlib.Path(str(self._output_path) + str(generated.path))
        os.makedirs(file_path.parent, exist_ok=True)
        if isinstance(generated.content, bytes):
            file_path.write_bytes(generated.content)
        else:
            file_path.write_text(generated.content, encoding='utf-8')
        return file_path

    def publish_static(self, static: StaticNode):
        import shutil
        import os
        from contextlib import suppress
        assert static is not None
        file_path = pathlib.Path(str(self._output_path) + str(static.path))
        os.makedirs(file_path.parent, exist_ok=True)

        with suppress(FileExistsError):
            # Symlink requires an absolute path
            source_path = os.path.abspath(static.src)
            try:
                os.symlink(source_path, file_path)
            # If we can't symlink for some reason (for instance,
            # Windows does not support symlinks by default, we try to
            # copy instead.
            except OSError:
                shutil.copyfile(source_path, file_path)

        return file_path


class TemplatePublisher(DefaultPublisher):
    def __init__(self, output_path: pathlib.Path, site: Site,
                 template_repository: TemplateRepository):
        super().__init__(output_path, site)
        self.__site_template_proxy = SiteTemplateProxy(self._site)
        self.__template_repository = template_repository

    def publish_document(self, document):
        assert document is not None
        return _publish_with_template(self._output_path, document,
                                      self._site,
                                      self.__site_template_proxy,
                                      self.__template_repository)

    def publish_index(self, index: IndexNode):
        assert index is not None
        return _publish_with_template(self._output_path, index,
                                      self._site,
                                      self.__site_template_proxy,
                                      self.__template_repository)



from setuptools import setup, find_packages
import liara

setup(
    name='Liara',
    version=liara.__version__,
    packages=find_packages(exclude=['*.test', 'test.*', '*.test.*']),
    install_requires=open('requirements.txt', 'r').readlines(),
    python_requires='>=3.6',

    author="Matthäus G. Chajdas",
    author_email="dev@anteru.net",
    description="Static page generator",
    long_description=open('README.md', 'r').read(),
    long_description_content_type="text/markdown",

    license="BSD",
    keywords=[],
    url="http://shelter13.net/projects/Liara",

    entry_points='''
        [console_scripts]
        liara=liara.cmdline:cli
    ''',
    setup_requires=['pytest-runner', 'sphinx'],
    tests_require=['pytest'],
    extras_require={
        'dev': [
            'flake8',
            'flake8-mypy',
        ]
    },

    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
    ]
)


from .nodes import (
    NodeKind
)
import pathlib
from .site import Site
from .template import TemplateRepository
from .publish import TemplatePublisher
from .cache import FilesystemCache
import logging
import webbrowser


class HttpServer:
    __log = logging.getLogger('liara.HttpServer')

    def __init__(self, site: Site, template_repository: TemplateRepository,
                 configuration, *, open_browser=True):
        self.__site = site
        self.__template_repository = template_repository
        self.__configuration = configuration
        self.__cache = FilesystemCache(pathlib.Path(
            self.__configuration['build.cache_directory']
        ))
        output_path = pathlib.Path(
            self.__configuration['output_directory'])
        self.__publisher = TemplatePublisher(output_path, self.__site,
                                             self.__template_repository)
        self.__open_browser = open_browser

    def _reload_template_paths(self):
        """Reload the template configuration.

        This ensures that any change to the template configuration is
        reflected in the template repository."""
        from .yaml import load_yaml
        template_configuration = pathlib.Path(self.__configuration['template'])
        configuration = load_yaml(template_configuration.open())
        self.__template_repository.update_paths(configuration['paths'])

    def _build_single_node(self, path: pathlib.PurePosixPath):
        """Build a single node.

        Build a single node on-demand. Special rules apply to make sure this is
        useful for actual work -- for instance, document/resource nodes
        are always rebuilt from scratch, and for documents, we also reload
        all templates."""
        from collections import namedtuple
        result = namedtuple('BuildResult', ['path', 'cache'])
        node = self.__site.get_node(path)

        if node is None:
            print(f'Could not find node for path: "{path}"')
            return (None, None,)

        # We always regenerate the content
        if node.kind in {NodeKind.Document, NodeKind.Resource}:
            node.reload()
            node.process(self.__cache)
            cache = False
        # We don't cache index nodes so templates get re-applied
        elif node.kind == NodeKind.Index:
            cache = False
        else:
            cache = True

        if node.kind == NodeKind.Document:
            self._reload_template_paths()

        return result(node.publish(self.__publisher), cache)

    def serve(self):
        """Serve the site with just-in-time processing.

        This does not build the whole site up-front, but rather builds nodes
        on demand. Nodes requiring templates are rebuilt from scratch every
        time to ensure they're up-to-date. Adding/removing nodes while
        serving will break the server, as it will not get notified and
        re-discover content."""
        import http.server

        class RequestHandler(http.server.BaseHTTPRequestHandler):
            def do_GET(self):
                path = pathlib.PurePosixPath(self.path)
                if path not in self.server.cache:
                    node_path, cache = \
                        self.server.http_server._build_single_node(path)

                    if node_path is None:
                        return

                    if cache:
                        self.server.cache[path] = node_path
                else:
                    node_path = self.server.cache[path]

                self.send_response(200)
                self.end_headers()
                self.wfile.write(node_path.open('rb').read())

            def log_message(self, f, *args):
                self.server.log.info(f, *args)

        server_address = ('', 8080)
        server = http.server.HTTPServer(server_address, RequestHandler)
        server.http_server = self
        server.log = self.__log
        server.cache = {}
        print('Listening on http://127.0.0.1:8080')
        if self.__open_browser:
            webbrowser.open('http://127.0.0.1:8080')
        server.serve_forever()
